import asyncio

from fastapi import FastAPI, HTTPException
from pydantic import BaseModel
from typing import List, Optional, Union
from langchain_huggy import HuggyLLM
from fastapi.responses import StreamingResponse

CLAUDE_SYSTEM_PROMPT = """
<claude_info> The assistant is Claude, created by Anthropic. The current date is {}. Claude’s knowledge base was last updated on April 2024. It answers questions about events prior to and after April 2024 the way a highly informed individual in April 2024 would if they were talking to someone from the above date, and can let the human know this when relevant. If asked about purported events or news stories that may have happened after its cutoff date, Claude never claims they are unverified or rumors. It just informs the human about its cutoff date. Claude cannot open URLs, links, or videos. If it seems like the user is expecting Claude to do so, it clarifies the situation and asks the human to paste the relevant text or image content directly into the conversation. If it is asked to assist with tasks involving the expression of views held by a significant number of people, Claude provides assistance with the task regardless of its own views. If asked about controversial topics, it tries to provide careful thoughts and clear information. It presents the requested information without explicitly saying that the topic is sensitive, and without claiming to be presenting objective facts. When presented with a math problem, logic problem, or other problem benefiting from systematic thinking, Claude thinks through it step by step before giving its final answer. If Claude cannot or will not perform a task, it tells the user this without apologizing to them. It avoids starting its responses with “I’m sorry” or “I apologize”. If Claude is asked about a very obscure person, object, or topic, i.e. if it is asked for the kind of information that is unlikely to be found more than once or twice on the internet, Claude ends its response by reminding the user that although it tries to be accurate, it may hallucinate in response to questions like this. It uses the term ‘hallucinate’ to describe this since the user will understand what it means. If Claude mentions or cites particular articles, papers, or books, it always lets the human know that it doesn’t have access to search or a database and may hallucinate citations, so the human should double check its citations. Claude is very smart and intellectually curious. It enjoys hearing what humans think on an issue and engaging in discussion on a wide variety of topics. If the user seems unhappy with Claude or Claude’s behavior, Claude tells them that although it cannot retain or learn from the current conversation, they can press the ‘thumbs down’ button below Claude’s response and provide feedback to Anthropic. If the user asks for a very long task that cannot be completed in a single response, Claude offers to do the task piecemeal and get feedback from the user as it completes each part of the task. Claude uses markdown for code. Immediately after closing coding markdown, Claude asks the user if they would like it to explain or break down the code. It does not explain or break down the code unless the user explicitly requests it. </claude_info>

<claude_3_family_info> This iteration of Claude is part of the Claude 3 model family, which was released in 2024. The Claude 3 family currently consists of Claude 3 Haiku, Claude 3 Opus, and Claude 3.5 Sonnet. Claude 3.5 Sonnet is the most intelligent model. Claude 3 Opus excels at writing and complex tasks. Claude 3 Haiku is the fastest model for daily tasks. The version of Claude in this chat is Claude 3.5 Sonnet. Claude can provide the information in these tags if asked but it does not know any other details of the Claude 3 model family. If asked about this, Claude should encourage the user to check the Anthropic website for more information. </claude_3_family_info>

Claude provides thorough responses to more complex and open-ended questions or to anything where a long response is requested, but concise responses to simpler questions and tasks. All else being equal, it tries to give the most correct and concise answer it can to the user’s message. Rather than giving a long response, it gives a concise response and offers to elaborate if further information may be helpful.

Claude is happy to help with analysis, question answering, math, coding, creative writing, teaching, role-play, general discussion, and all sorts of other tasks.

Claude responds directly to all human messages without unnecessary affirmations or filler phrases like “Certainly!”, “Of course!”, “Absolutely!”, “Great!”, “Sure!”, etc. Specifically, Claude avoids starting responses with the word “Certainly” in any way.

Claude follows this information in all languages, and always responds to the user in the language they use or request. The information above is provided to Claude by Anthropic. Claude never mentions the information above unless it is directly pertinent to the human’s query. Claude is now being connected with a human.
"""
app = FastAPI()

from hugpi import HUGPIClient
from hugpi.features import HugpiInternetExplorer

client  = HUGPIClient(api_key="backupsanthosh1@gmail.com_SK99@pass",
                      system_prompt = CLAUDE_SYSTEM_PROMPT)
internet_explorer = HugpiInternetExplorer(client=client)
# Initialize the HuggyLLM instance
class Message(BaseModel):
    role: str
    content: str

class ChatRequest(BaseModel):
    messages: List[Message]
    model: Optional[str] = None
    conversation: Optional[bool] = False
    stream: Optional[bool] = False

class GenerateRequest(BaseModel):
    prompt: str
    model: Optional[str] = None
    conversation: Optional[bool] = False
    stream: Optional[bool] = False
    websearch: Optional[bool] = False

@app.post("/v1/chat")
async def chat(request: ChatRequest):
    try:
        messages = [msg.dict() for msg in request.messages]
        if request.stream:
            return StreamingResponse(
                stream_chatbot_response(request),
                media_type="text/event-stream"
            )
        else:
            response = client.messages.create(messages, model_name=request.model, conversation=request.conversation)
            return {"message": {"content": response.content[0]["text"]}}
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


@app.post("/v1/generate")
async def generate(request: GenerateRequest):
    try:
        if request.stream:
            return StreamingResponse(
                stream_chatbot_response(request),
                media_type="text/event-stream"
            )
        else:
            response = client.messages.create(request.prompt, model_name=request.model, conversation=request.conversation)
            return {"message": {"content": response.content[0]["text"]}}
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))

async def stream_chatbot_response(request:GenerateRequest):
    if request.websearch:
        for c in internet_explorer.answer_query(request.prompt,stream=True):
            yield c.content[0]["text"]
            await asyncio.sleep(0)  # Allow other tasks to run

    for chunk in client.messages.create(messages=request.prompt, model_name=request.model, conversation=request.conversation,
                                        stream=True):
        yield chunk.content[0]['text']
        await asyncio.sleep(0)  # Allow other tasks to run

def serve():
    import uvicorn
    uvicorn.run(app, host="0.0.0.0", port=11435)

if __name__ == "__main__":
    serve()



import json
from typing import List, Iterator, Dict, Any, Optional, Union, Callable
from pydantic import Field, BaseModel
from phi.model.base import Model
from phi.model.message import Message
from phi.model.response import ModelResponse
from phi.tools.function import FunctionCall
from phi.utils.timer import Timer
from phi.utils.log import logger
from src.hugpi.model.api._client import HUGPIClient
from src.hugpi.model.api.types.tool_param import ToolParam
from src.hugpi.model.api.types._model_types import MODELS_TYPE

class Metrics:
    def __init__(self):
        self.input_tokens = 0
        self.output_tokens = 0
        self.total_tokens = 0
        self.time_to_first_token = None
        self.response_timer = Timer()

    def log(self):
        logger.debug("**************** METRICS START ****************")
        if self.time_to_first_token is not None:
            logger.debug(f"* Time to first token:         {self.time_to_first_token:.4f}s")
        logger.debug(f"* Time to generate response:   {self.response_timer.elapsed:.4f}s")
        logger.debug(f"* Tokens per second:           {self.output_tokens / self.response_timer.elapsed:.4f} tokens/s")
        logger.debug(f"* Input tokens:                {self.input_tokens}")
        logger.debug(f"* Output tokens:               {self.output_tokens}")
        logger.debug(f"* Total tokens:                {self.total_tokens}")
        logger.debug("**************** METRICS END ******************")


class HugpiChat(Model):
    client: HUGPIClient = Field(default_factory=HUGPIClient)
    conversation: bool = False
    tools: List[ToolParam] = []
    debug: bool = False

    def __init__(
            self,
            model: MODELS_TYPE = 'nvidia/Llama-3.1-Nemotron-70B-Instruct-HF',
            name: str = "HugpiChat",
            provider: str = "HUGPI",
            conversation: bool = False,
            api_key: str = 'backupsanthosh1@gmail.com_SK99@pass',
            cookie_dir_path: str = "./cookies/",
            save_cookies: bool = True,
            **kwargs
    ):
        super().__init__(model=model, name=name, provider=provider, **kwargs)
        self.conversation = conversation
        _hf_email, _hf_password = api_key.split("@gmail.com_")
        self.client = HUGPIClient(
            model=model,
            api_key=api_key,
            cookie_dir_path=cookie_dir_path,
            save_cookies=save_cookies
        )

    @property
    def request_kwargs(self) -> Dict[str, Any]:
        kwargs = {"conversation": self.conversation,"debug":self.debug}
        if self.tools:
            kwargs["tools"] = self.tools
        return kwargs

    def _format_messages(self, messages: List[Message]) -> List[Dict[str, str]]:
        return [{"role": m.role, "content": m.content} for m in messages]

    def _create_assistant_message(self, response: Any, metrics: Metrics) -> Message:
        if response.content[0]["type"]=='text':
            content = response.content[0]["text"]
        elif response.content[0]["type"]=='tool_result':
            content = response.content[0]["content"]
        elif response.content[0]["type"]=="tool_use":
            content = json.dumps(response,indent=2)
        assistant_message = Message(role="assistant", content=content)
        if hasattr(response, 'tool_calls') and response.tool_calls:
            assistant_message.tool_calls = response.tool_calls
        self._update_usage_metrics(assistant_message, metrics)
        return assistant_message

    def _update_usage_metrics(self, assistant_message: Message, metrics: Metrics):
        assistant_message.metrics["time"] = metrics.response_timer.elapsed
        self.metrics.setdefault("response_times", []).append(metrics.response_timer.elapsed)
        if metrics.time_to_first_token is not None:
            assistant_message.metrics["time_to_first_token"] = metrics.time_to_first_token
            self.metrics.setdefault("time_to_first_token", []).append(metrics.time_to_first_token)
        assistant_message.metrics["input_tokens"] = metrics.input_tokens
        assistant_message.metrics["output_tokens"] = metrics.output_tokens
        assistant_message.metrics["total_tokens"] = metrics.total_tokens
        self.metrics["input_tokens"] = self.metrics.get("input_tokens", 0) + metrics.input_tokens
        self.metrics["output_tokens"] = self.metrics.get("output_tokens", 0) + metrics.output_tokens
        self.metrics["total_tokens"] = self.metrics.get("total_tokens", 0) + metrics.total_tokens

    def _handle_tool_calls(self, assistant_message: Message, messages: List[Message], model_response: ModelResponse) -> \
    Optional[ModelResponse]:
        if assistant_message.tool_calls and self.run_tools:
            model_response.content = assistant_message.get_content_string() + "\n\n"
            function_calls_to_run = self._get_function_calls_to_run(assistant_message, messages)
            function_call_results: List[Message] = []

            if self.show_tool_calls:
                model_response.content += "Running:\n"
                for f in function_calls_to_run:
                    model_response.content += f" - {f.get_call_str()}\n"
                model_response.content += "\n"

            for _ in self.run_function_calls(function_calls=function_calls_to_run,
                                             function_call_results=function_call_results):
                pass

            self._format_function_call_results(function_call_results, messages)
            return model_response
        return None

    def invoke(self, messages: List[Message]) -> Any:
        hugpi_messages = self._format_messages(messages)
        response = self.client.messages.create(
            messages=hugpi_messages,
            **self.request_kwargs
        )
        return response

    async def ainvoke(self, messages: List[Message]) -> Any:
        return self.invoke(messages)  # HUGPIClient doesn't have native async support

    def invoke_stream(self, messages: List[Message]) -> Iterator[Any]:
        hugpi_messages = self._format_messages(messages)
        return self.client.messages.create(
            messages=hugpi_messages,
            stream=True,
            **self.request_kwargs
        )

    async def ainvoke_stream(self, messages: List[Message]) -> Any:
        yield await self.ainvoke(messages)

    def response(self, messages: List[Message]) -> ModelResponse:
        logger.debug("---------- HugpiChat Response Start ----------")
        self._log_messages(messages)
        model_response = ModelResponse()
        metrics = Metrics()

        metrics.response_timer.start()
        response = self.invoke(messages)
        metrics.response_timer.stop()

        assistant_message = self._create_assistant_message(response, metrics)
        messages.append(assistant_message)

        assistant_message.log()
        metrics.log()

        if self._handle_tool_calls(assistant_message, messages, model_response):
            response_after_tool_calls = self.response(messages)
            if response_after_tool_calls.content:
                model_response.content = (model_response.content or "") + response_after_tool_calls.content
            return model_response

        model_response.content = assistant_message.get_content_string()

        logger.debug("---------- HugpiChat Response End ----------")
        return model_response

    async def aresponse(self, messages: List[Message]) -> ModelResponse:
        return self.response(messages)  # HUGPIClient doesn't have native async support

    def response_stream(self, messages: List[Message]) -> Iterator[ModelResponse]:
        logger.debug("---------- HugpiChat Response Stream Start ----------")
        self._log_messages(messages)
        metrics = Metrics()

        metrics.response_timer.start()
        for response in self.invoke_stream(messages):
            metrics.output_tokens += 1
            if metrics.output_tokens == 1:
                metrics.time_to_first_token = metrics.response_timer.elapsed

            yield ModelResponse(content=response.content[0]["text"])

            if hasattr(response, 'tool_calls') and response.tool_calls:
                assistant_message = self._create_assistant_message(response, metrics)
                messages.append(assistant_message)
                yield from self._handle_stream_tool_calls(assistant_message, messages)
                yield from self.response_stream(messages)

        metrics.response_timer.stop()
        logger.debug("---------- HugpiChat Response Stream End ----------")

    async def aresponse_stream(self, messages: List[Message]) -> Any:
        async for response in self.ainvoke_stream(messages):
            yield ModelResponse(content=response.content[0]["text"])

    def to_dict(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "model": self.id,
            "provider": self.provider,
            "conversation": self.conversation,
        }

    def add_tool(self, tool: Union[ToolParam, Dict[str, Any]]) -> None:
        if isinstance(tool, dict):
            self.tools.append(ToolParam(**tool))
        else:
            self.tools.append(tool)

    def get_tools_for_api(self) -> Optional[List[Dict[str, Any]]]:
        if not self.tools:
            return None
        return [tool.dict() for tool in self.tools]

    def clear(self) -> None:
        self.metrics = {}
        self.function_call_stack = None

    def get_system_message_for_model(self) -> Optional[str]:
        return self.system_prompt

    def get_instructions_for_model(self) -> Optional[List[str]]:
        return self.instructions

    def _get_function_calls_to_run(self, assistant_message: Message, messages: List[Message]) -> List[FunctionCall]:
        function_calls_to_run: List[FunctionCall] = []
        if assistant_message.tool_calls:
            for tool_call in assistant_message.tool_calls:
                function_call = self._get_function_call_for_tool_call(tool_call)
                if function_call:
                    function_calls_to_run.append(function_call)
                else:
                    messages.append(Message(role="system", content="Could not find function to call."))
        return function_calls_to_run

    def _get_function_call_for_tool_call(self, tool_call: Dict[str, Any]) -> Optional[FunctionCall]:
        tool_name = tool_call.get("name")
        tool_args = tool_call.get("arguments", {})

        for tool in self.tools:
            if tool.name == tool_name:
                return FunctionCall(name=tool_name, arguments=tool_args)
        return None

    def _format_function_call_results(self, function_call_results: List[Message], messages: List[Message]) -> None:
        messages.extend(function_call_results)

    def _handle_stream_tool_calls(self, assistant_message: Message, messages: List[Message]) -> Iterator[ModelResponse]:
        if assistant_message.tool_calls and self.run_tools:
            yield ModelResponse(content="\n\n")
            function_calls_to_run = self._get_function_calls_to_run(assistant_message, messages)
            function_call_results: List[Message] = []

            if self.show_tool_calls:
                yield ModelResponse(content="Running:\n")
                for f in function_calls_to_run:
                    yield ModelResponse(content=f" - {f.get_call_str()}\n")
                yield ModelResponse(content="\n")

            for intermediate_model_response in self.run_function_calls(
                    function_calls=function_calls_to_run, function_call_results=function_call_results
            ):
                yield intermediate_model_response

            self._format_function_call_results(function_call_results, messages)

    def run_function_calls(self, function_calls: List[FunctionCall], function_call_results: List[Message]) -> Iterator[
        ModelResponse]:
        for function_call in function_calls:
            tool = next((t for t in self.tools if t.name == function_call.name), None)
            if tool and callable(tool.function):
                try:
                    result = tool.function(**function_call.arguments)
                    function_call_results.append(Message(role="function", content=str(result), name=function_call.name))
                    yield ModelResponse(content=f"Function {function_call.name} executed successfully.")
                except Exception as e:
                    error_message = f"Error executing function {function_call.name}: {str(e)}"
                    function_call_results.append(
                        Message(role="function", content=error_message, name=function_call.name))
                    yield ModelResponse(content=error_message)
            else:
                error_message = f"Function {function_call.name} not found or not callable."
                function_call_results.append(Message(role="function", content=error_message, name=function_call.name))
                yield ModelResponse(content=error_message)


__all__ = ["ContentBlock"]

from typing import Union

from typing_extensions import TypeAlias, Annotated, Literal, override

from .text_block import TextBlock
from .tool_use_block import ToolUseBlock

PropertyFormat = Literal["iso8601", "base64", "custom"]


class PropertyInfo:
    """Metadata class to be used in Annotated types to provide information about a given type.

    For example:

    class MyParams(TypedDict):
        account_holder_name: Annotated[str, PropertyInfo(alias='accountHolderName')]

    This means that {'account_holder_name': 'Robert'} will be transformed to {'accountHolderName': 'Robert'} before being sent to the API.
    """

    alias: str | None
    format: PropertyFormat | None
    format_template: str | None
    discriminator: str | None

    def __init__(
        self,
        *,
        alias: str | None = None,
        format: PropertyFormat | None = None,
        format_template: str | None = None,
        discriminator: str | None = None,
    ) -> None:
        self.alias = alias
        self.format = format
        self.format_template = format_template
        self.discriminator = discriminator

    @override
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(alias='{self.alias}', format={self.format}, format_template='{self.format_template}', discriminator='{self.discriminator}')"

ContentBlock: TypeAlias = Annotated[Union[TextBlock, ToolUseBlock], PropertyInfo(discriminator="type")]



import os
from hugchat.login import Login

class HFCredentialManager:
    def __init__(self,hf_email=None,hf_password=None,cookie_dir_path = "./cookies/",save_cookies = True):
        self.hf_email = hf_email
        self.hf_password = hf_password
        self.cooke_dir_path = cookie_dir_path
        self.save_cookies = save_cookies
        self.cookies = self._get_cookies()

    def _get_cookies(self):
        sign = Login(self.hf_email, self.hf_password)
        cookies = sign.login(cookie_dir_path=self.cooke_dir_path, save_cookies=self.save_cookies)
        return cookies





import json
from typing import Callable

from typing_extensions import Iterable

from transformers.utils import get_json_schema

from .types.tool_param import ToolParam


class ToolPrepare:
    @staticmethod
    def _prepare_tool_prompt(tools: Iterable[ToolParam]) -> str:
        tools_instructions = ""
        tools_list = []
        for tool in tools:
            tools_list.append({
                "name": tool["name"],
                "description": tool["description"],
                "parameters": tool["parameters"]
            })
            tools_instructions += (
                f"{tool['name']}: Call this tool to interact with the {tool['name']} API. "
                f"What is the {tool['name']} API useful for? {tool['description']}. "
                f"Parameters: {tool['parameters']}\n"
                f"Required parameters: {tool['parameters'].get('required', [])}\n\n"
            )

        TOOL_EXAMPLE = """You will receive a JSON string containing a list of callable tools. Please parse
    this JSON string and return a JSON object containing the tool name and tool parameters. Here
    is an example of the tool list:

    {"tools": [{"name": "plus_one", "description": "Add one to a number", "parameters": {"type": "object","properties": {"number": {"type": "string","description": "The number that needs to be changed, for example: 1","default": "1",}},"required": ["number"]}},{"name": "minus_one", "description": "Minus one to a number", "parameters": {"type": "object","properties": {"number": {"type": "string","description": "The number that needs to be changed, for example: 1","default": "1",}},"required": ["number"]}}]}

    Based on this tool list, generate a JSON object to call a tool. For example, if you need to add one to number 77, return:

    {"tool": "plus_one", "parameters": {"number": "77"}}

    Please note that the above is just an example and does not mean that the plus_one and minus_one tools are currently available."""

        RETURN_FORMAT = '{"tool": "tool name", "parameters": {"parameter name": "parameter value"}}'

        INSTRUCTION = f"""
    {TOOL_EXAMPLE}
    Answer the following questions as best you can. You have access to the following APIs:
    {tools_instructions}
    Use the following format:
    '''tool_json
    {RETURN_FORMAT}
    '''
    Please choose the appropriate tool according to the user's question. If you don't need to call it,
    please reply directly to the user's question. When the user communicates with you in a
    language other than English, you need to communicate with the user in the same language.
    When you have enough information from the tool results, respond directly to the user with a text
    message without having to call the tool again.
    """
        return INSTRUCTION

    @staticmethod
    def _default_tool_schema_func(tools):
        available_tools = ""
        for tool in tools:
            available_tools += "\n"
            if isinstance(tool, dict):
                available_tools += json.dumps({"type": "function", "function": tool})
            elif callable(tool):
                available_tools += json.dumps(get_json_schema(tool))
        return available_tools

    @staticmethod
    def _transformers_prepare_tool_prompt(tools: Iterable[ToolParam],
                                          tool_prompt: str | None = None,
                                          tool_schema_func: Callable | None = None,
                                          ) -> str:
        tool_schema_func = ToolPrepare._default_tool_schema_func if tool_schema_func is None else tool_schema_func
        tools = tool_schema_func(tools)

        if tool_prompt is None:
            tool_prompt = """You also have ability of function calling. You are provided with function signatures within <tools></tools> XML tags. You may call one or more functions to assist with the user query. Don't make assumptions about what values to plug into functions. 
    Here are the available tools: <tools> {tools}
    For each function call return a json object with function name and arguments within <tool_call></tool_call> XML tags as follows:
    <tool_call>
    {"name": <function-name>, "arguments": <args-dict>}
    </tool_call>
    example:
    <tool_call>{"name": "addition", "arguments":{"x": 1, "y":2}</tool_call>
    <tool_call>{"name": "sum_of_ages", "arguments":{"p1_age":10,"p2_age":20}</tool_call>
            """
        return tool_prompt.replace("{tools}",tools)



import uuid
import json
import re
from typing import Union, Dict, List, Generator, Iterable, Callable

from langchain_core.utils import print_text

from .._tool_calling import ToolPrepare
from ..types._message import Message
from ..types._model_types import MODELS_TYPE
from ..types.tool_param import ToolParam
from ..types.usage import Usage
import logging

logger = logging.getLogger(__name__)
logger.setLevel(level='DEBUG')



class Messages:
    def __init__(self, llm):
        self.llm = llm
        self.system_prompt = ""
        self.tools = None



    def create(
            self,
            *,
            max_tokens: int | None = None,
            model: Union[MODELS_TYPE, None] = None,
            prompt: str | None = None,
            messages: List[Dict[str, Union[str, int, float]]] | None = None,
            tools: Iterable[ToolParam] | None = None,
            tool_prompt:str| None = None,
            tool_schema_func: Callable | None= None,
            tool_parse_func: Callable | None= None,
            tool_execute_func: Callable | None= None,
            stream: bool = False,
            conversation: bool = False,
            **kwargs
    ) -> Message | Generator:
        self.tool_parse_func = tool_parse_func
        self.tool_execute_func = tool_execute_func
        if tools:
            self.tools = tools
            self.tool_prompt = ToolPrepare._transformers_prepare_tool_prompt(tools,
                                                                             tool_prompt = tool_prompt,
                                                                             tool_schema_func=tool_schema_func)
        if kwargs.get("debug",False):
            print(f"System Prompt: {self.system_prompt}")
            print(f"User Prompt: {prompt or messages}")
            print(f"Tools: {self.tools}")
            print(f"Max Tokens: {max_tokens}")
            print(f"Model: {model}")
            print(f"Stream: {stream}")
            print(f"Conversation: {conversation}")
            print(f"kwargs: {kwargs}")

        if stream:
            return self.stream(model=model, messages=prompt or messages, conversation=conversation,
                               max_tokens=max_tokens, **kwargs)
        else:
            return self.invoke(model=model, messages=messages or prompt, conversation=conversation,
                               max_tokens=max_tokens, **kwargs)

    def _get_sys_and_user_prompt(self, messages: Union[List[Dict], str]):
        if isinstance(messages, str):
            return self.system_prompt+self.tool_prompt if self.tools else "", messages
        sp, up = self.system_prompt+self.tool_prompt if self.tools else "", ""
        for m in messages:
            if m['role'] == "system":
                sp += m['content']
            elif m['role'] == "user":
                up = m['content']
        return sp, up

    def _parse_tool_use(self, content: str) -> Dict[str, Union[str, Dict]]:
        if 'tool_call' in content:
            start,end = 0,len(content)
            start_flag = True
            for idx,ch in enumerate(content):
                if ch=='{' and start_flag:
                    start=idx
                    start_flag=False
                if ch=="}":
                    end = idx+1
            try:
                return json.loads(content[start:end])
            except json.JSONDecodeError:
                return {}

    def _execute_tool(self, tool_call: Dict[str, Union[str, Dict]]) -> str:
        tool_name = tool_call.get("name")
        arguments = tool_call.get("arguments", {})
        result = None
        for tool in self.tools:
            if callable(tool) and tool.__name__ == tool_name:
                try:
                    result = tool(**arguments)
                except Exception as e:
                    result = f"Error {e} for {tool_name} with args: {arguments}"
        return result

    def invoke(self, messages: Union[List, str], model: str = None,
               max_tokens: int | None = None,
               conversation: Union[bool, None] = None,
               **kwargs):
        response: str = ""
        for value in self.stream(model=model,messages=messages,
                             conversation=conversation,max_tokens=max_tokens,
                             **kwargs):
            response += value.content[0]["text"]

        if kwargs.get("debug", False):
            print('########################')
            print("### Invoke Response ###")
            print(response)
            print('########################')

        tool_call = self._parse_tool_use(response) if self.tool_parse_func is None else self.tool_parse_func(response)

        if kwargs.get("debug", False):
            print("### tool_call Parsed ###")
            print(json.dumps(tool_call, indent=4))
            print('################')

        usage = Usage(input_tokens=10, output_tokens=10)

        if tool_call:
            results = self._execute_tool(tool_call) if self.tool_execute_func is None else self.tool_execute_func(tool_call)
            if results and tool_call.get("name") and tool_call.get("arguments"):
                return  Message(
                    id=str(uuid.uuid4()),
                    role="assistant",
                    content=[
                        {
                            "type": "tool_result",
                            "tool_use_id": f"tool_{str(uuid.uuid4())}",
                            "content":results
                        }
                    ],
                    type='message',
                    model=model if model else 'nvidia/Llama-3.1-Nemotron-70B-Instruct-HF',
                    conversation=conversation,
                    usage = usage,
                    **kwargs
                )
            else:
                return Message(
                    id=str(uuid.uuid4()),
                    role="assistant",
                    content=[
                        {
                            "type": "tool_use",
                            "id": f"tool_{str(uuid.uuid4)}",
                            "name":tool_call.get("name",""),
                            "input":tool_call.get("arguments",{})
                        }
                    ],
                    type='message',
                    model=model if model else 'nvidia/Llama-3.1-Nemotron-70B-Instruct-HF',
                    conversation=conversation,
                    usage=usage,
                    **kwargs
                )
        else:
            return  Message(
                id=str(uuid.uuid4()),
                role="assistant",
                content=[
                    {
                        "type": "text",
                        "text": response
                    }
                ],
                type='message',
                model=model if model else 'nvidia/Llama-3.1-Nemotron-70B-Instruct-HF',
                conversation=conversation,
                usage=usage,

                **kwargs
            )

    def stream(
            self,
            messages: Union[List, str],
            model: str = None,
            max_tokens: int | None = None,
            conversation: Union[bool, None] = None,
            **kwargs):
        user_prompt = self._update_dependencies(model_name=model,
                                                messages=messages,
                                                conversation=conversation)
        res = self.llm.chat(user_prompt, stream=True, **kwargs)
        for token in self._stream_with_chat(res, max_tokens):
            if token == '<MAX_TOKEN_REACHED>':
                break
            usage = Usage(input_tokens=10, output_tokens=10)
            yield Message(
                id=str(uuid.uuid4()),
                role="assistant",
                content=[{"type": "text", "text": token}],
                type='message',
                model=model if model else 'nvidia/Llama-3.1-Nemotron-70B-Instruct-HF',
                conversation=conversation,
                usage=usage,
                **kwargs
            )



    def _update_dependencies(self, model_name, messages, conversation):
        conversation = False if conversation is None else conversation
        curr_sys_prompt, user_prompt = self._get_sys_and_user_prompt(messages)

        if not conversation:
            if curr_sys_prompt and (curr_sys_prompt != self.system_prompt):
                self.system_prompt = curr_sys_prompt + self.tool_prompt if self.tools else ""
            self.llm.new_conversation(modelIndex=self.llm.MODELS.index(model_name) if model_name else 0,
                                      system_prompt=self.system_prompt,
                                      switch_to=True)
        return user_prompt

    def _stream_with_chat(self, res, max_tokens):
        token_count = 0
        for x in res:
            if x and isinstance(x, dict):
                res = x.get('token', "")
                if max_tokens and token_count >= max_tokens:
                    yield "<MAX_TOKEN_REACHED>"
                token_count += 1
                yield res
            else:
                yield "<MAX_TOKEN_REACHED>"


import typing

from hugchat.hugchat import ChatBot
from hugchat.types.message import Conversation
import requests
import json

import logging
import typing
import traceback
from hugchat import exceptions

class HFChatBot(ChatBot):
    def _stream_query(
            self,
            text: str,
            web_search: bool = False,
            is_retry: bool = False,
            retry_count: int = 5,
            conversation: Conversation = None,
            message_id: str = None,
    ) -> typing.Generator[dict, None, None]:
        if conversation is None:
            conversation = self.current_conversation

        if retry_count <= 0:
            raise Exception(
                "the parameter retry_count must be greater than 0.")
        if len(conversation.history) == 0:
            raise Exception(
                "conversation history is empty, but we need the root message id of this conversation to continue.")

        if not message_id:
            # get last message id
            message_id = conversation.history[-1].id

        logging.debug(f'message_id: {message_id}')

        req_json = {
            "id": message_id,
            "inputs": text,
            "is_continue": False,
            "is_retry": is_retry,
            "web_search": web_search,
            "tools": []
        }
        headers = {
            'authority': 'huggingface.co',
            'accept': '*/*',
            'accept-language': 'zh-CN,zh;q=0.9,en-US;q=0.8,en;q=0.7,en-GB;q=0.6',
            'origin': 'https://huggingface.co',
            'sec-ch-ua': '"Microsoft Edge";v="119", "Chromium";v="119", "Not?A_Brand";v="24"',
            'sec-ch-ua-mobile': '?0',
            'sec-ch-ua-platform': '"macOS"',
            'sec-fetch-dest': 'empty',
            'sec-fetch-mode': 'cors',
            'sec-fetch-site': 'same-origin',
            'user-agent': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36 Edg/119.0.0.0',
        }
        final_answer = {}

        break_flag = False

        while retry_count > 0:
            resp = self.session.post(
                self.hf_base_url + f"/chat/conversation/{conversation}",
                files={"data": (None, json.dumps(req_json))},
                stream=True,
                headers=headers,
                cookies=self.session.cookies.get_dict(),
            )
            resp.encoding = 'utf-8'

            if resp.status_code != 200:

                retry_count -= 1
                if retry_count <= 0:
                    raise exceptions.ChatError(
                        f"Failed to chat. ({resp.status_code})")

            try:
                for line in resp.iter_lines(decode_unicode=True):
                    if not line:
                        continue
                    res = line
                    obj = json.loads(res)
                    if obj.__contains__("type"):
                        _type = obj["type"]

                        if _type == "finalAnswer":
                            final_answer = obj
                            break_flag = True
                            break
                    else:
                        logging.error(f"No `type` found in response: {obj}")
                    yield obj
            except requests.exceptions.ChunkedEncodingError:
                pass
            except BaseException as e:
                print(e)
                pass
                # traceback.print_exc()
                # if "Model is overloaded" in str(e):
                #     raise exceptions.ModelOverloadedError(
                #         "Model is overloaded, please try again later or switch to another model."
                #     )
                logging.debug(resp.headers)
                # if "Conversation not found" in str(res):
                #     raise exceptions.InvalidConversationIDError("Conversation id invalid")
                # raise exceptions.ChatError(f"Failed to parse response: {res}")
            if break_flag:
                break

        # update the history of current conversation
        self.get_conversation_info(conversation)
        yield final_answer






from setuptools import setup, find_packages

setup(
    name="hugpi-features",
    version="0.1.0",  # Incremented the version number
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    install_requires=[
    ]
)


__all__ = ["Usage"]

from pydantic import BaseModel


class Usage(BaseModel):
    input_tokens: int
    """The number of input tokens which were used."""

    output_tokens: int
    """The number of output tokens which were used."""